/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/master.hh"

#include "config.h"

#include <boost/program_options.hpp>

#include <exception>
#include <fstream>
#include <iostream>

namespace CdnskeyProcessor {
namespace Cfg {

namespace {

constexpr const char* to_string(Master::Mode mode)
{
    switch (mode)
    {
        case Master::Mode::consumer:
            return "consumer";
        case Master::Mode::producer:
            return "producer";
    }
    struct InvalidMode : std::exception
    {
        const char* what() const noexcept override { return "invalid mode"; }
    };
    throw InvalidMode{};
}

}//namespace CdnskeyProcessor::Cfg::{anonymous}

std::ostream& operator<<(std::ostream& out, Master::Mode mode)
{
    return out << to_string(mode);
}

void validate(
        boost::any& v,
        const std::vector<std::string>& values,
        Master::Mode*,
        int)
{
    // Make sure no previous assignment to 'v' was made.
    boost::program_options::validators::check_first_occurrence(v);
    // Extract the first string from 'values'. If there is more than
    // one string, it's an error, and exception will be thrown.
    const auto& s = boost::program_options::validators::get_single_string(values);
    if (s == to_string(Master::Mode::consumer))
    {
        v = boost::any{Master::Mode::consumer};
    }
    else if (s == to_string(Master::Mode::producer))
    {
        v = boost::any{Master::Mode::producer};
    }
    else
    {
        throw boost::program_options::invalid_option_value{"invalid mode"};
    }
}

namespace {

struct GeneralArgs
{
    std::string config_file;
    Master::Mode running_mode;
};

boost::program_options::options_description get_options(GeneralArgs& general_args)
{
    namespace po = boost::program_options;
    po::options_description options{"General options", program_options_line_length};
    options.add_options()
        ("conf,c", po::value<std::string>(&general_args.config_file)->required(),
         "configuration file")
        ("help,h",
         "display this help message")
        ("running_mode,m", po::value<Master::Mode>(&general_args.running_mode)->default_value(Master::Mode::producer),
         "consumer reads results and stores them into database, producer sends tasks to workers");
    return options;
}

boost::program_options::options_description get_options(Master::Amqp::Broker& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Broker options", program_options_line_length};
    log_options.add_options()
        ("amqp.broker.hostname",
         po::value<std::string>(&cfg.hostname)->default_value("localhost"),
         "RabbitMQ broker hostname")
        ("amqp.broker.port", po::value<std::uint16_t>(&cfg.port), "RabbitMQ broker port")
        ("amqp.broker.vhost",
         po::value<std::string>(&cfg.vhost)->default_value("cdnskey-processor-dev"),
         "RabbitMQ broker virtual host")
        ("amqp.broker.username",
         po::value<std::string>(&cfg.username)->default_value("guest"), "RabbitMQ username")
        ("amqp.broker.password",
         po::value<std::string>(&cfg.password)->default_value("guest"), "RabbitMQ password");
    return log_options;
}

boost::program_options::options_description get_options(Master::Amqp::Exchange& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Exchange options", program_options_line_length};
    log_options.add_options()
        ("amqp.exchange.insecure_fanout",
         po::value<std::string>(&cfg.insecure_fanout)->default_value("exchange_insecure_fanout"),
         "AMQP Exchange for insecure(fanout) scan requests")
        ("amqp.exchange.secure",
         po::value<std::string>(&cfg.secure)->default_value("exchange_secure"),
         "AMQP Exchange for secure scan requests")
        ("amqp.exchange.response",
         po::value<std::string>(&cfg.response)->default_value("exchange_response"),
         "AMQP Exchange for workers' responses");
    return log_options;
}

boost::program_options::options_description get_options(Master::Amqp::Queue& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Queue options", program_options_line_length};
    log_options.add_options()
        ("amqp.queue.results",
         po::value<std::string>(&cfg.results)->default_value("queue_results"),
         "AMQP Queue for scan results (connected to 'response' exchange")
        ("amqp.queue.insecure_prefix",
         po::value<std::string>(&cfg.insecure_prefix)->default_value("queue_insecure_"),
         "AMQP Queue name prefix for worker to consume insecure tasks (worker's id appended)")
        ("amqp.queue.secure",
         po::value<std::string>(&cfg.secure)->default_value("queue_secure"),
         "AMQP Queue name for worker to consume secure tasks");
    return log_options;
}

boost::program_options::options_description get_options(Master::Amqp& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"AMQP General options", program_options_line_length};
    options.add_options()
        ("amqp.prefetch_count",
         po::value<std::uint16_t>(&cfg.prefetch_count)->default_value(1),
         "defines the max number of unacknowledged deliveries that are permitted on a channel");
    options.add(get_options(cfg.broker))
           .add(get_options(cfg.exchange))
           .add(get_options(cfg.queue));
    return options;
}

}//namespace CdnskeyProcessor::Cfg::{anonymous}

}//namespace CdnskeyProcessor::Cfg
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Cfg;

Master CdnskeyProcessor::Cfg::get_cfg(
        int argc,
        char** argv,
        void(*set_up_logging)(const Log&),
        void(*set_up_database)(const Database&))
{
    const auto general_args = [&]()
    {
        GeneralArgs general_args;
        auto all_options = get_options(general_args);
        boost::program_options::variables_map vm;
        boost::program_options::store(boost::program_options::parse_command_line(argc, argv, all_options), vm);
        if (vm.count("help"))
        {
            Master cfg;
            Log log_cfg;
            Database db_cfg;
            all_options.add(get_options(log_cfg))
                       .add(get_options(db_cfg))
                       .add(get_options(cfg.amqp));
            std::cout << all_options << std::flush;
            throw Master::Help{};
        }
        notify(vm);
        return general_args;
    }();
    auto cfg_file = std::ifstream{general_args.config_file};
    if (!cfg_file.is_open())
    {
        struct UnableToOpen : std::exception
        {
            const char* what() const noexcept override { return "unable to open config file"; }
        };
        throw UnableToOpen{};
    }
    Master cfg;
    Log log_cfg;
    Database db_cfg;
    boost::program_options::options_description all_options{"cdnskey-processor-master options"};
    all_options.add(get_options(log_cfg))
               .add(get_options(db_cfg))
               .add(get_options(cfg.amqp));
    boost::program_options::variables_map vm;
    boost::program_options::store(boost::program_options::parse_config_file(cfg_file, all_options), vm);
    boost::program_options::notify(vm);
    if (set_up_logging != nullptr)
    {
        set_up_logging(log_cfg);
    }
    if (set_up_database != nullptr)
    {
        set_up_database(db_cfg);
    }
    cfg.running_mode = general_args.running_mode;
    return cfg;
}
