/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/master.hh"
#include "src/database.hh"
#include "src/internal/amqp/boost_asio_handler.hh"
#include "src/internal/amqp/consume.hh"
#include "src/internal/amqp/exchange.hh"
#include "src/internal/amqp/publish.hh"
#include "src/internal/amqp/queue.hh"
#include "src/internal/amqp/reliable_tcp_channel.hh"
#include "src/internal/amqp/tcp_connection.hh"
#include "src/internal/scan_result.hh"
#include "src/internal/timer.hh"
#include "src/scan_task.hh"
#include "src/storage.hh"

#include <liblog/liblog.hh>

#include <boost/asio/signal_set.hpp>

#include <amqpcpp.h>
#include <amqpcpp/libboostasio.h>

#include <atomic>
#include <csignal>
#include <cstdlib>
#include <iostream>
#include <thread>
#include <tuple>
#include <utility>

#define NOTHROW_LOG_TRACE(msg, ...) do { try { LIBLOG_TRACE(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_DEBUG(msg, ...) do { try { LIBLOG_DEBUG(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_INFO(msg, ...) do { try { LIBLOG_INFO(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_WARNING(msg, ...) do { try { LIBLOG_WARNING(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_ERROR(msg, ...) do { try { LIBLOG_ERROR(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_CRITICAL(msg, ...) do { try { LIBLOG_CRITICAL(msg, ## __VA_ARGS__); } catch (...) {} } while (false)

namespace CdnskeyProcessor {

class MasterConsumer
{
public:
    explicit MasterConsumer(
            Internal::Timer&,
            const Cfg::Master&);
    ~MasterConsumer();
    void stop_requested();
private:
    struct LogCtx
    {
        explicit LogCtx()
            : ctx{"consumer"}
        {
            LIBLOG_DEBUG("context entered");
        }
        ~LogCtx()
        {
            try
            {
                LIBLOG_DEBUG("leaving context");
            }
            catch (...) { }
        }
        LibLog::SetContext ctx;
    } log_ctx_;
    Cfg::Master cfg_;
    std::unique_ptr<AMQP::TcpConnection> make_amqp_connection();
    void on_connection_established();
    void on_connection_broke_down(const AMQP::TcpConnection*);
    void on_queue_results_delivered(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered);
    Internal::Timer& timer_;
    std::unique_ptr<AMQP::TcpConnection> connection_;
    std::unique_ptr<AMQP::TcpChannel> channel_;
    Internal::Amqp::Queue queue_results_;
};

class MasterProducer
{
public:
    explicit MasterProducer(
            Internal::Timer&,
            const Cfg::Master&);
    ~MasterProducer();
    void stop_requested();
private:
    struct LogCtx
    {
        explicit LogCtx()
            : ctx{"producer"}
        {
            LIBLOG_DEBUG("context entered");
        }
        ~LogCtx()
        {
            try
            {
                LIBLOG_DEBUG("leaving context");
            }
            catch (...) { }
        }
        LibLog::SetContext ctx;
    } log_ctx_;
    Cfg::Master cfg_;
    std::unique_ptr<AMQP::TcpConnection> make_amqp_connection();
    void on_connection_established();
    void on_connection_broke_down(const AMQP::TcpConnection*);
    void step(std::chrono::steady_clock::time_point now);
    using NumberOfSent = LibStrong::TypeWithSkills<
            std::size_t,
            struct NumberOfSentTag_,
            LibStrong::Skill::Countable,
            LibStrong::Skill::Streamable>;
    NumberOfSent send_scans();
    void on_scan_batch_delivered(ScanBatchId);
    Internal::Timer& timer_;
    std::unique_ptr<AMQP::TcpConnection> connection_;
    std::unique_ptr<Internal::Amqp::ReliableTcpChannel> channel_;
    Internal::Amqp::FanoutExchange exchange_insecure_fanout_;
    Internal::Amqp::FanoutExchange exchange_secure_;
};

namespace {

volatile std::sig_atomic_t occuring_signal = -1;

}//namespace CdnskeyProcessor::{anonymous}

}//namespace CdnskeyProcessor

int main(int argc, char* argv[])
{
    try
    {
        const auto cfg = [argc, argv]()
        {
            try
            {
                return CdnskeyProcessor::Cfg::get_cfg(
                        argc,
                        argv,
                        CdnskeyProcessor::Cfg::set_up_logging,
                        CdnskeyProcessor::set_up_database);
            }
            catch (const CdnskeyProcessor::Cfg::Master::Help&)
            {
                std::exit(EXIT_SUCCESS);
            }
            catch (const std::exception& e)
            {
                std::cerr << "std::exception caught: " << e.what() << std::endl;
            }
            catch (...)
            {
                std::cerr << "unknown exception caught" << std::endl;
            }
            std::exit(EXIT_FAILURE);
        }();
        CdnskeyProcessor::Internal::Timer timer{1};
        boost::asio::signal_set signals{timer.service(), SIGINT, SIGHUP, SIGQUIT};
        signals.add(SIGTERM);
        switch (cfg.running_mode)
        {
            case CdnskeyProcessor::Cfg::Master::Mode::consumer:
            {
                CdnskeyProcessor::MasterConsumer consumer{timer, cfg};
                signals.async_wait([&consumer](const boost::system::error_code& error, int signal_number)
                {
                    if (!error)
                    {
                        CdnskeyProcessor::occuring_signal = signal_number;
                        consumer.stop_requested();
                    }
                });
                timer.loop();
                return EXIT_SUCCESS;
            }
            case CdnskeyProcessor::Cfg::Master::Mode::producer:
            {
                CdnskeyProcessor::MasterProducer producer{timer, cfg};
                signals.async_wait([&producer](const boost::system::error_code& error, int signal_number)
                {
                    if (!error)
                    {
                        CdnskeyProcessor::occuring_signal = signal_number;
                        producer.stop_requested();
                    }
                });
                timer.loop();
                return EXIT_SUCCESS;
            }
        }
        LIBLOG_ERROR("invalid master configuration: unknown running mode");
        return EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("std::exception caught: {}", e.what());
        return EXIT_FAILURE;
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown exception caught");
        return EXIT_FAILURE;
    }
}

namespace CdnskeyProcessor {

namespace {

std::string get_signal()
{
    if (occuring_signal == SIGINT)
    {
        return "SIGINT";
    }
    if (occuring_signal == SIGHUP)
    {
        return "SIGHUP";
    }
    if (occuring_signal == SIGQUIT)
    {
        return "SIGQUIT";
    }
    if (occuring_signal == SIGTERM)
    {
        return "SIGTERM";
    }
    return "unknown " + std::to_string(occuring_signal);
}

AMQP::Address make_address(const Cfg::Master::Amqp::Broker& cfg)
{
    return AMQP::Address{cfg.hostname, cfg.port, {cfg.username, cfg.password}, cfg.vhost};
}

template <typename Rep, typename Period>
void set_expiration(AMQP::Envelope& envelope, std::chrono::duration<Rep, Period> ttl)
{
    const auto ttl_ms = std::chrono::duration_cast<std::chrono::milliseconds>(ttl);
    envelope.setExpiration(std::to_string(ttl_ms.count()));
}

}//namespace CdnskeyProcessor::{anonymous}

MasterConsumer::MasterConsumer(
        Internal::Timer& timer,
        const Cfg::Master& cfg)
    : log_ctx_{},
      cfg_{cfg},
      timer_{timer},
      connection_{make_amqp_connection()},
      channel_{nullptr},
      queue_results_{cfg.amqp.queue.results}
{
    static constexpr auto delay_to_establish_connection = std::chrono::minutes{1};
    timer_.register_callback(Internal::Timer::TimePoint::clock::now() + delay_to_establish_connection, [](auto&&) {});
}

MasterConsumer::~MasterConsumer()
{
    NOTHROW_LOG_INFO("consumer done by signal {}", get_signal());
}

std::unique_ptr<AMQP::TcpConnection> MasterConsumer::make_amqp_connection()
{
    return Internal::Amqp::make_tcp_connection(
            Internal::Amqp::make_boost_asio_handler(
                    timer_.service(),
                    [this](AMQP::TcpConnection*)
                    {
                        this->on_connection_established();
                    },
                    [this](AMQP::TcpConnection* conn, const char* message)
                    {
                        NOTHROW_LOG_ERROR("connection to the RabbitMQ failed: {}", message);
                        this->on_connection_broke_down(conn);
                    },
                    [this](AMQP::TcpConnection* conn)
                    {
                        this->on_connection_broke_down(conn);
                    }),
            make_address(cfg_.amqp.broker));
}

void MasterConsumer::on_connection_established()
{
    NOTHROW_LOG_INFO("consumer established connection to the RabbitMQ");
    channel_ = std::make_unique<AMQP::TcpChannel>(connection_.get());
    channel_->setQos(cfg_.amqp.prefetch_count);
    static constexpr int no_flags = 0;
    consume(*channel_, queue_results_, std::string{}, no_flags, AMQP::Table{})
        .onReceived([this](const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
        {
            try
            {
                NOTHROW_LOG_TRACE("results queue consumed: {} bytes", message.bodySize());
                this->on_queue_results_delivered(message, delivery_tag, redelivered);
            }
            catch (const std::exception& e)
            {
                NOTHROW_LOG_ERROR("exception caught by results queue handling: {}", e.what());
            }
            catch (...)
            {
                NOTHROW_LOG_ERROR("unknow exception caught by results queue handling");
            }
        })
        .onError([](const char* message)
        {
            NOTHROW_LOG_ERROR("results queue consume failed: {}", message);
        })
        .onFinalize([]()
        {
            NOTHROW_LOG_INFO("results queue consume finalized");
        });
}

void MasterConsumer::on_connection_broke_down(const AMQP::TcpConnection* broken_conn)
{
    if (connection_.get() != broken_conn)
    {
        NOTHROW_LOG_INFO("loss of this connection has already been treated");
        return;
    }
    NOTHROW_LOG_INFO("consumer lost connection to the RabbitMQ; wait 10s and reestablish it");
    channel_.reset();
    connection_.reset();
    using namespace std::chrono_literals;
    static constexpr auto delay_between_connection_attempts = 10s;
    std::this_thread::sleep_for(delay_between_connection_attempts);
    connection_ = make_amqp_connection();
}

void MasterConsumer::stop_requested()
{
    timer_.service().stop();
}

void MasterConsumer::on_queue_results_delivered(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
{
    const auto scan_result = deserialize<ScanResultCollection>(message.body(), message.bodySize());
    if (redelivered)
    {
        LIBLOG_INFO("scan result {} redelivered", *scan_result.scan_batch_id);
    }
    else
    {
        LIBLOG_TRACE("scan result {} delivered", *scan_result.scan_batch_id);
    }
    auto tx = make_rw_transaction();
    batch_result_received(tx, scan_result);
    commit(std::move(tx));
    const bool acknowledged = channel_->ack(delivery_tag);
    if (!acknowledged)
    {
        LIBLOG_INFO("queue results did not acknowledge receiving of a result");
    }
}

MasterProducer::MasterProducer(
        Internal::Timer& timer,
        const Cfg::Master& cfg)
    : log_ctx_{},
      cfg_{cfg},
      timer_{timer},
      connection_{make_amqp_connection()},
      channel_{nullptr},
      exchange_insecure_fanout_{Internal::Amqp::ExchangeName{cfg_.amqp.exchange.insecure_fanout}},
      exchange_secure_{Internal::Amqp::ExchangeName{cfg_.amqp.exchange.secure}}
{
    static constexpr auto delay_to_establish_connection = std::chrono::minutes{1};
    timer_.register_callback(Internal::Timer::TimePoint::clock::now() + delay_to_establish_connection, [](auto&&) {});
}

MasterProducer::~MasterProducer()
{
    NOTHROW_LOG_INFO("producer done by signal {}", get_signal());
}

std::unique_ptr<AMQP::TcpConnection> MasterProducer::make_amqp_connection()
{
    return Internal::Amqp::make_tcp_connection(
            Internal::Amqp::make_boost_asio_handler(
                    timer_.service(),
                    [this](AMQP::TcpConnection*)
                    {
                        this->on_connection_established();
                    },
                    [this](AMQP::TcpConnection* conn, const char* message)
                    {
                        NOTHROW_LOG_ERROR("connection to the RabbitMQ failed: {}", message);
                        this->on_connection_broke_down(conn);
                    },
                    [this](AMQP::TcpConnection* conn)
                    {
                        this->on_connection_broke_down(conn);
                    }),
            make_address(cfg_.amqp.broker));
}

void MasterProducer::on_connection_established()
{
    NOTHROW_LOG_INFO("producer established connection to the RabbitMQ");
    channel_ = std::make_unique<Internal::Amqp::ReliableTcpChannel>(std::make_unique<AMQP::TcpChannel>(connection_.get()));
    (**channel_).onError([](const char* msg) { NOTHROW_LOG_INFO("error on reliable channel: {}", msg); });
    static bool timer_is_running = false;
    if (!timer_is_running)
    {
        step(std::chrono::steady_clock::now());
        timer_is_running = true;
    }
}

void MasterProducer::on_connection_broke_down(const AMQP::TcpConnection* broken_conn)
{
    if (connection_.get() != broken_conn)
    {
        NOTHROW_LOG_INFO("loss of this connection has already been treated");
        return;
    }
    NOTHROW_LOG_INFO("producer lost connection to the RabbitMQ; wait 10s and reestablish it");
    channel_.reset();
    connection_.reset();
    using namespace std::chrono_literals;
    static constexpr auto delay_between_connection_attempts = 10s;
    std::this_thread::sleep_for(delay_between_connection_attempts);
    connection_ = make_amqp_connection();
}

void MasterProducer::stop_requested()
{
    timer_.service().stop();
}

void MasterProducer::step(std::chrono::steady_clock::time_point now)
{
    static constexpr auto step_period = std::chrono::minutes{10};
    timer_.register_callback(now + step_period, [this](auto&& now) { step(now); });
    const auto scans = send_scans();
    if (NumberOfSent{0ull} < scans)
    {
        LIBLOG_INFO("{} scans sent", *scans);
    }
}

MasterProducer::NumberOfSent MasterProducer::send_scans()
{
    NumberOfSent total_number_of_sent = NumberOfSent{0ull};
    auto db_conn = make_db_connection();
    while (true)
    {
        const auto data = [&]()
        {
            auto tx = LibPg::PgRoTransaction{std::move(db_conn)};
            const auto batch = get_insecure_batch_to_solve(tx);
            db_conn = commit(std::move(tx));
            const auto number_of_sent = NumberOfSent{batch.jobs.size()};
            if (number_of_sent == NumberOfSent{0ull})
            {
                return std::make_tuple(ScanBatchId{0ll}, std::string{});
            }
            total_number_of_sent += number_of_sent;
            return std::make_tuple(batch.batch_id, serialize(batch));
        }();
        if (std::get<std::string>(data).empty())
        {
            break;
        }
        AMQP::Envelope envelope{std::get<std::string>(data).c_str(), std::get<std::string>(data).size()};
        set_expiration(envelope, std::chrono::hours{22});
        envelope.setPersistent(true);
        static constexpr int no_flags = 0;
        auto& published = publish(**channel_, exchange_insecure_fanout_, envelope, no_flags);
        {
            auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
            batch_sent(tx, std::get<ScanBatchId>(data));
            db_conn = commit(std::move(tx));
        }
        const auto& batch_id = std::get<ScanBatchId>(data);
        published.onAck([this, batch_id]()
                {
                    this->on_scan_batch_delivered(batch_id);
                })
                .onSuccess([batch_id]() { NOTHROW_LOG_INFO("insecure scan batch {} published", *batch_id); })
                .onError([batch_id](const char* msg) { NOTHROW_LOG_INFO("insecure scan batch {} publishing failed: {}", *batch_id, msg); })
                .onFinalize([batch_id]() { NOTHROW_LOG_INFO("insecure scan batch {} publishing finalized", *batch_id); });
    }
    while (true)
    {
        const auto data = [&]()
        {
            auto tx = make_rw_transaction();
            const auto batch = get_secure_batch_to_solve(tx);
            commit(std::move(tx));
            const auto number_of_sent = NumberOfSent{batch.jobs.size()};
            if (number_of_sent == NumberOfSent{0ull})
            {
                return std::make_tuple(ScanBatchId{0ll}, std::string{});
            }
            total_number_of_sent += number_of_sent;
            return std::make_tuple(batch.batch_id, serialize(batch));
        }();
        if (std::get<std::string>(data).empty())
        {
            break;
        }
        AMQP::Envelope envelope{std::get<std::string>(data).c_str(), std::get<std::string>(data).size()};
        set_expiration(envelope, std::chrono::hours{22});
        envelope.setPersistent(true);
        static constexpr int no_flags = 0;
        auto& published = publish(**channel_, exchange_secure_, envelope, no_flags);
        {
            auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
            batch_sent(tx, std::get<ScanBatchId>(data));
            db_conn = commit(std::move(tx));
        }
        const auto& batch_id = std::get<ScanBatchId>(data);
        published.onAck([this, batch_id]()
                {
                    this->on_scan_batch_delivered(batch_id);
                })
                .onSuccess([batch_id]() { NOTHROW_LOG_INFO("secure scan batch {} published", *batch_id); })
                .onError([batch_id](const char* msg) { NOTHROW_LOG_INFO("secure scan batch {} publishing failed: {}", *batch_id, msg); })
                .onFinalize([batch_id]() { NOTHROW_LOG_INFO("secure scan batch {} publishing finalized", *batch_id); });
    }
    return total_number_of_sent;
}

void MasterProducer::on_scan_batch_delivered(ScanBatchId batch_id)
{
    LIBLOG_INFO("scan batch {} delivered to all workers", *batch_id);
    auto tx = make_rw_transaction();
    batch_started(tx, batch_id);
    commit(std::move(tx));
}

}//namespace CdnskeyProcessor
